################################################################################
###### Implementation for experiments with the genome data (Fig 3 in the paper).
###### Usage: python2 Genome.py -f trash_glycerol.wig
################################################################################


import sys
import numpy
import scipy.stats
import math
import hmm_tools


wigPath = ""
if "-f" in sys.argv:
    wigPath = sys.argv[sys.argv.index("-f") +1]
else:
    print("Usage: print %s -f <file> [-gff <GFF3 File>]" % sys.argv[0])
    sys.exit()

gffPath = ""
if "-gff" in sys.argv:
    gffPath = sys.argv[sys.argv.index("-gff") +1]

N = 4
defaultStates = True
if "-n" in sys.argv: 
    if "-p" in sys.argv and "-a" in sys.argv:
        N = int(sys.argv[sys.argv.index("-n") +1])
        defaultStates = False
    else:
        print("Usage: print %s -f <file> [-gff <GFF3 File>]\n" % sys.argv[0])
        print('Note: If specifying the number of states you must also specify the parameters (-p), \
and transition probabilities (-a). Please see the instructions for more information.\n')
        sys.exit()


defaultParameters = True
if "-p" in sys.argv:
    PARAM = numpy.fromstring(sys.argv[sys.argv.index("-p")+1], sep=",")
    defaultParameters = False

defaultTransition = True
if "-a" in sys.argv:
    PROB = numpy.array(numpy.mat(sys.argv[sys.argv.index("-a") +1]))
    defaultTransition = False




# Get Gene Annotation from GFF File
hash = {}; rv2name = {}
if gffPath:
    hash = hmm_tools.hash_genes(gffPath)
    rv2name = hmm_tools.get_name(gffPath)



# Read in read count data
pos_list = []
C = []
for line in open(wigPath):
    if line.startswith("#"): continue
    if line.startswith("variable"): continue
    tmp = line.split()
    pos = int(tmp[0]); reads = int(tmp[1]);
    pos_list.append(int(pos))
    C.append(reads+1)

O = numpy.array(C) # Reads



# If default states, set known labels
if defaultStates:
    label= {0:"ES", 1:"GD", 2:"NE",3:"GA"}
else:
    label = dict([(s, "State-%d" % s) for s in range(0,N)])


reads = O-1
reads_nz = sorted(reads[reads !=0 ])
size = len(reads_nz)
mean_r = numpy.average(reads_nz[:int(0.95 * size)])



# If parameters haven't changed, set them as specified in paper
if defaultParameters:
    mu = numpy.array([1/0.99, 0.01 * mean_r + 2,  mean_r, mean_r*5.0])
    L = 1.0/mu
    label= {0:"ES", 1:"GD", 2:"NE",3:"GA"}
else:
    L = PARAM



B = [] # Emission Probability Distributions
for i in range(N):
    B.append(scipy.stats.geom(L[i]).pmf)


pins = hmm_tools.calculate_pins(O-1)
pins_obs = sum([1 for rd in O if rd >=2])/float(len(O))
# IF Default Transition, calculate as specified in paper
if defaultTransition:
    pnon = 1.0 - pins
    pnon_obs = 1.0 - pins_obs

    for r in range(100):
        if pnon ** r < 0.01: break

    A = numpy.zeros((N,N))
    a = math.log1p(-B[int(N/2)](1)**r + numpy.finfo(float).eps)
    b = r*math.log(B[int(N/2)](1) + numpy.finfo(float).eps) + math.log(1.0/3)
    for i in range(N):
        A[i] = [b]*N
        A[i][i] = a

else:
    A = numpy.log(PROB)
    r = -1

PI = numpy.zeros(N) # Initial state distribution
PI[0] = 0.7; PI[1:] = 0.3/(N-1);


print('Number of observations', len(O))
################
#### VITERBI ###
(Q_opt, delta, Q) = hmm_tools.viterbi(A, B, PI, O, scaling=True, discrete=False)
################


###############
# PEEK SEARCH #
##############
actual_states = ["ES", "GD", "NE", "GA"]
label_inverse_map = {"ES":0, "GD":1, "NE":2, "GA":3}
adjustment = 64


#### Get emission and transition probabilities for each model.

# First get the log of start state distribution
log_start_p = {}
for state in actual_states:
    log_start_p[state] = numpy.log(PI[label_inverse_map[state]] + numpy.finfo(float).eps) + adjustment

    
# Next get the log of transition probabilities
log_transition_probs = {}
for state_i in actual_states:
    log_transition_probs[state_i] = {}
    for state_j in actual_states:
        log_transition_probs[state_i][state_j] = A[label_inverse_map[state_i]][label_inverse_map[state_j]] + adjustment  
        #Note log is already computed on transitions, so we don't do log(A). 


#### Get the log-probability and fraction of agreement (Fig. 3).             
Viterbi_L = hmm_tools.compute_log_probability(actual_states, Q_opt, O, log_start_p, log_transition_probs, B, label_inverse_map, adjustment)
   

for L in [1, 3, 5, 7, 9, 11, 13, 15, 17, 19]: 
    Q_opt_Peek_states = hmm_tools.PeekSearch(B, label_inverse_map, adjustment, actual_states, log_transition_probs, log_start_p, obs = O, L=L)
    Q_opt_Peek = [label_inverse_map[state] for state in Q_opt_Peek_states]

    assert len(Q_opt) == len(Q_opt_Peek)
    match = numpy.sum([1.0 for j in range(len(Q_opt)) if Q_opt[j] == Q_opt_Peek[j]])/len(Q_opt)
    
    Peek_L = hmm_tools.compute_log_probability(actual_states, Q_opt_Peek, O, log_start_p, log_transition_probs, B, label_inverse_map, adjustment)
    print('Viterbi log-probability : ' + str(Viterbi_L) + '. Peek log-probability for L = ' + str(L) + ': ' + \
          str(Peek_L) + '. Match fraction : ' + str(match))
    











